import { strictEqual } from './equality';
import useEntity from './useEntity';
import alias from '../plugins/alias';
import { onInit, onSet } from '../tools/inspector';
import { getStore, isStoreEnabled } from '../tools/store';

function entity(initialValue, pluginsOrAlias = []) {
  if (initialValue === undefined) throw new Error('Entity requires an initial value.');
  const plugins = typeof pluginsOrAlias === 'string' ? [alias(pluginsOrAlias)] : pluginsOrAlias;
  if (!(plugins instanceof Array)) throw new Error('Invalid plug-ins array.');
  const entityBuild = {
    _value: undefined,
    _subscribers: new Set()
  };
  entityBuild.subscribe = createSubscribe(entityBuild);
  entityBuild.get = createGet(entityBuild);
  entityBuild.set = createSet(entityBuild);
  entityBuild.init = createInit(entityBuild, initialValue);
  entityBuild.use = createHook(entityBuild);
  const newEntity = entityBuild;
  applyPlugins(newEntity, plugins);
  newEntity.init();
  if (isStoreEnabled()) getStore().add(newEntity);
  return newEntity;
}

export default entity;

function createSubscribe(entity) {
  return subscriberFn => {
    entity._subscribers.add(subscriberFn);

    return () => {
      entity._subscribers.delete(subscriberFn);
    };
  };
}

function createGet(entity) {
  return () => entity._value;
}

function createSet(entity) {
  return (valueOrUpdaterFn, alias) => {
    if (typeof valueOrUpdaterFn === 'function') valueOrUpdaterFn = valueOrUpdaterFn(entity._value);
    entity._value = valueOrUpdaterFn;

    entity._subscribers.forEach(cb => cb(entity._value));

    if (alias !== '@@DEVTOOLS') onSet(entity, alias ?? '<anonymous>');
  };
}

let nextId = 1;

function createInit(entity, initialValue) {
  return () => {
    if (!entity.name) entity.name = `entity${nextId++}`;
    if (initialValue instanceof Promise) initialValue.then(value => setTimeout(() => entity.set(value, '@@ASYNC_INIT')));else entity._value = initialValue;
    onInit(entity);
  };
}

function createHook(entity) {
  return (transform = v => v, equality = strictEqual) => useEntity(entity, transform, equality);
}

function applyPlugins(entity, plugins) {
  plugins.forEach(plugin => {
    if (typeof plugin !== 'object') throw new Error('Invalid plug-in');

    function overrideMethod(method) {
      const createOverride = plugin[method];

      if (typeof createOverride === 'function') {
        const override = createOverride(entity[method], entity);
        if (typeof override !== 'function') throw new Error(`Invalid override for '${method}' in plug-in.`);
        entity[method] = override;
      }
    }

    Object.keys(plugin).forEach(method => overrideMethod(method));
  });
}